// ATMEGA8

#define F_CPU	8000000
#define BAUD	19200
#define MYUBR	F_CPU / 16 / BAUD - 1

#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/delay.h>

#define KBD_C		PD3
#define KBD_C_DDR	DDRD
#define KBD_C_PIN	PIND
#define KBD_C_PORT	PORTD

#define KBD_D		PC4
#define KBD_D_DDR	DDRC
#define KBD_D_PIN	PINC
#define KBD_D_PORT	PORTC

#define CRT_C		PD2
#define CRT_C_DDR	DDRD
#define CRT_C_PIN	PIND
#define CRT_C_PORT	PORTD

#define CRT_D		PC5
#define CRT_D_DDR	DDRC
#define CRT_D_PIN	PINC
#define CRT_D_PORT	PORTC

#define KBD_DIR_RECEIVE		0
#define KBD_DIR_SEND		1

#define KBD_LED_SCROLL		1
#define KBD_LED_NUM		2
#define KBD_LED_CAPS		4

volatile unsigned int CRT_CHAR;
volatile unsigned int CRT_DATA;

volatile unsigned char KBD_CHAR;
volatile unsigned char KBD_BIT_COUNT;
volatile unsigned int KBD_DATA;
volatile unsigned char KBD_DIR;
volatile unsigned char KBD_LED;

unsigned char CRT_MAP[] =
	{
	0x0D, 0x09, 0x00, 0x00,							// <TABULATOR>
	0x0E, '^' , '' , 0x00,
	0x15, 'q' , 'Q' , '@' ,
	0x16, '1' , '!' , 0x00,
	0x1A, 'z' , 'Z' , 0x00,
	0x1B, 's' , 'S' , 0x00,
	0x1C, 'a' , 'A' , 0x00,
	0x1D, 'w' , 'W' , 0x00,
	0x1E, '2' , '"' , '' ,
	0x21, 'c' , 'C' , 0x00,
	0x22, 'x' , 'X' , 0x00,
	0x23, 'd' , 'D' , 0x00,
	0x24, 'e' , 'E' , '' ,
	0x25, '4' , '$' , 0x00,
	0x26, '3' , '$' , '' ,
	0x29, ' ' , ' ' , 0x00,
	0x2A, 'v' , 'V' , 0x00,
	0x2B, 'f' , 'F' , 0x00,
	0x2C, 't' , 'T' , 0x00,
	0x2D, 'r' , 'R' , 0x00,
	0x2E, '5' , '%' , 0x00,
	0x31, 'n' , 'N' , 0x00,
	0x32, 'b' , 'B' , 0x00,
	0x33, 'h' , 'H' , 0x00,
	0x34, 'g' , 'G' , 0x00,
	0x35, 'y' , 'Y' , 0x00,
	0x36, '6' , '&' , 0x00,
	0x3A, 'm' , 'M' , '' ,
	0x3B, 'j' , 'J' , 0x00,
	0x3C, 'u' , 'U' , 0x00,
	0x3D, '7' , '/' , '{' ,
	0x3E, '8' , '(' , '[' ,
	0x41, ',' , ';' , 0x00,
	0x42, 'k' , 'K' , 0x00,
	0x43, 'i' , 'I' , 0x00,
	0x44, 'o' , 'O' , '}' ,
	0x45, '0' , '=' , 0x00,
	0x46, '9' , ')' , ']' ,
	0x49, '.' , ':' , 0x00,
	0x4A, '-' , '_' , 0x00,
	0x4B, 'l' , 'L' , 0x00,
	0x4C, '' , '' , 0x00,
	0x4D, 'p' , 'P' , 0x00,
	0x4E, '' , '?' , '\\',
	0x52, '' , '' , 0x00,
	0x54, '' , '' , 0x00,
	0x55, '' , '`' , 0x00,
	0x5A, 0x0D, 0x0D, 0x00,							// <ENTER>
	0x5B, '+' , '*' , '~' ,
	0x5D, '#' , '\'', 0,00,
	0x61, '<' , '>' , '|' ,
	0x66, 0x08, 0x08, 0x00,							// <BACKSPACE>
	0x69, '1' , 0x00, 0x00,
	0x6B, '4' , 0x00, 0x00,
	0x6C, '7' , 0x00, 0x00,
	0x70, '0' , 0x00, 0x00,
	0x71, ',' , 0x00, 0x00,
	0x72, '2' , 0x00, 0x00,
	0x73, '5' , 0x00, 0x00,
	0x74, '6' , 0x00, 0x00,
	0x75, '8' , 0x00, 0x00,
	0x76, 0x27, 0x27, 0x00,							// <ESCAPE>
	0x79, '+' , 0x00, 0x00,
	0x7A, '3' , 0x00, 0x00,
	0x7B, '-' , 0x00, 0x00,
	0x7C, '*' , 0x00, 0x00,
	0x7D, '9' , 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00
	};

void usart_write_char(char value)
	{
	while((UCSRA & (1 << UDRE)) == 0x00);

	UDR = value;
	}

void usart_write_hex_char(unsigned char value)
	{
	unsigned char iData = 0x00;

	iData = ((value & 0xF0) >> 4);

	usart_write_char((iData > 0x09) ? (iData + 0x37) : (iData + 0x30));

	iData = ((value & 0x0F) >> 0);

	usart_write_char((iData > 0x09) ? (iData + 0x37) : (iData + 0x30));
	}

void usart_write_hex_int(unsigned int value)
	{
	usart_write_hex_char((value & 0xFF00) >> 8);
	usart_write_hex_char((value & 0x00FF) >> 0);
	}

void usart_write_hex_long(unsigned long value)
	{
	usart_write_hex_int((value & 0xFFFF0000) >> 8);
	usart_write_hex_int((value & 0x0000FFFF) >> 0);
	}

void usart_write_string(char * string)
	{
	while(* string)
		{
		usart_write_char(* string ++);
		}
	}

unsigned int create_frame(unsigned char keycode)
	{
	unsigned int frame = 0x0000;
	unsigned char parity = 0x00;

	for(unsigned char bit = 0; bit < 8; bit = bit + 1)
		{
		if(keycode & 0x01)
			{
			frame = frame | 0x0400;
			parity = parity + 1;
			}

		keycode = (keycode >> 1);
		frame = (frame >> 1);
		}

	frame = (frame >> 1);

	if(parity & 0x01)
		{
		frame = frame | 0x0400;						// parity 0 stop 1
		}
	else
		{
		frame = frame | 0x0600;						// parity 1 stop 1
		}

	return(frame);
	}

void crt_read_frame(void)
	{
	CRT_C_DDR = CRT_C_DDR | (1 << CRT_C);					// CLOCK OUTPUT
	CRT_D_DDR = CRT_D_DDR & ~ (1 << CRT_D);					// DATA INPUT

	for(unsigned char bit = 0; bit < 11; bit = bit + 1)
		{
		CRT_DATA = (CRT_DATA >> 1);

		if(CRT_D_PIN & (1 << CRT_D))
			{
			CRT_DATA = (CRT_DATA | 0x0400);
			}

		_delay_us(20);
		CRT_C_PORT = CRT_C_PORT & ~ (1 << CRT_C);			// CLOCK OFF
		_delay_us(20);
		CRT_C_PORT = CRT_C_PORT | (1 << CRT_C);				// CLOCK ON
		_delay_us(20);
		}
	}

unsigned char crt_get_state(void)
	{
	CRT_C_DDR = CRT_C_DDR & ~ (1 << CRT_C);					// CLOCK INPUT
	CRT_D_DDR = CRT_D_DDR & ~ (1 << CRT_D);					// DATA INPUT

	if((CRT_C_PIN & (1 << CRT_C)) == (1 << CRT_C))
		{
		if((CRT_D_PIN & (1 << CRT_D)) == (1 << CRT_D))
			{
			return(3);						// HOST READY
			}
		else
			{
			return(2);						// HOST GOT DATA
			}
		}
	else
		{
		if((CRT_D_PIN & (1 << CRT_D)) == (1 << CRT_D))
			{
			return(1);						// HOST BUSY
			}
		else
			{
			return(0);						// HOST RESET
			}
		}
	}

void crt_write_frame(unsigned int frame)
	{
	while(crt_get_state() != 0x03);

	CRT_C_DDR = CRT_C_DDR | (1 << CRT_C);					// CLOCK OUTPUT
	CRT_D_DDR = CRT_D_DDR | (1 << CRT_D);					// DATA OUTPUT

	for(unsigned char bit = 0; bit < 11; bit = bit + 1)
		{
		if(frame & 0x0001)
			{
			CRT_D_PORT = CRT_D_PORT | (1 << CRT_D);			// DATA ON
			}
		else
			{
			CRT_D_PORT = CRT_D_PORT & ~ (1 << CRT_D);		// DATA OFF
			}

		frame = (frame >> 1);

		_delay_us(20);
		CRT_C_PORT = CRT_C_PORT & ~ (1 << CRT_C);			// CLOCK OFF
		_delay_us(20);
		CRT_C_PORT = CRT_C_PORT | (1 << CRT_C);				// CLOCK ON
		_delay_us(20);
		}

	CRT_D_PORT = CRT_D_PORT | (1 << CRT_D);					// DATA ON !!!
	}

void crt_write_scancode(unsigned char iChar)
	{
	unsigned int iData = create_frame(iChar);

	crt_write_frame(iData);
	}

void crt_write_char(unsigned char iChar)
	{
	unsigned int crt_map_index = 0x0000;

	while(1)
		{
		if(CRT_MAP[crt_map_index + 0] == 0x00)
			{
			break;
			}

		if(CRT_MAP[crt_map_index + 1] == iChar)
			{
			iChar = CRT_MAP[crt_map_index + 0];

			// button - press
			crt_write_scancode(iChar);

			// button - release
			crt_write_scancode(0xF0);
			crt_write_scancode(iChar);

			break;
			}

		if(CRT_MAP[crt_map_index + 2] == iChar)
			{
			iChar = CRT_MAP[crt_map_index + 0];

			// shift left - press
			crt_write_scancode(0x12);

			// button - press
			crt_write_scancode(iChar);

			// button - release
			crt_write_scancode(0xF0);
			crt_write_scancode(iChar);

			// shift left - release
			crt_write_scancode(0xF0);
			crt_write_scancode(0x12);

			break;
			}

		if(CRT_MAP[crt_map_index + 3] == iChar)
			{
			iChar = CRT_MAP[crt_map_index + 0];

			// altgr - press
			crt_write_scancode(0xE0);
			crt_write_scancode(0x11);

			// button - press
			crt_write_scancode(iChar);

			// button - release
			crt_write_scancode(0xF0);
			crt_write_scancode(iChar);

			// altgr - release
			crt_write_scancode(0xE0);
			crt_write_scancode(0xF0);
			crt_write_scancode(0x11);

			break;
			}

		crt_map_index = (crt_map_index + 4);
		}
	}

void crt_write_string(char * string)
	{
	while(* string)
		{
		crt_write_char(* string ++);
		}
	}

ISR(SIG_INTERRUPT1)
	{
	if(KBD_DIR == KBD_DIR_RECEIVE)						// RECEIVE FROM KBD
		{
		if(KBD_C_PIN & (1 << KBD_C))					// The rising edge of INT1
			{
			MCUCR = MCUCR & ~ (1 << ISC10);				// The falling edge of INT1 generates an interrupt request.

			KBD_BIT_COUNT = (KBD_BIT_COUNT - 1);

			if(KBD_BIT_COUNT == 0)
				{
				KBD_CHAR = ((KBD_DATA >> 1) & 0xFF);

				crt_write_frame(KBD_DATA); 			// Daten an CRT weiterleiten

				KBD_BIT_COUNT = 11;
				}
			}
		else								// The falling edge of INT1
			{
			MCUCR = MCUCR | (1 << ISC10);				// The rising edge of INT1 generates an interrupt request.

			KBD_DATA = (KBD_DATA >> 1);

			if(KBD_D_PIN & (1 << KBD_D))
				{
				KBD_DATA = (KBD_DATA | 0x0400);
				}
			}
		}
	else									// SEND TO KBD
		{
		if(KBD_C_PIN & (1 << KBD_C))					// The rising edge of INT1
			{
			MCUCR = MCUCR & ~ (1 << ISC10);				// The falling edge of INT1 generates an interrupt request.

			KBD_BIT_COUNT = (KBD_BIT_COUNT - 1);

			if(KBD_BIT_COUNT == 0)
				{
				KBD_BIT_COUNT = 11;

				KBD_DIR = KBD_DIR_RECEIVE;

				KBD_D_PORT = KBD_D_PORT | (1 << KBD_D);		// DATA ON
				}
			}
		else								// The falling edge of INT1
			{
			MCUCR = MCUCR | (1 << ISC10);				// The rising edge of INT1 generates an interrupt request.

			if(KBD_DATA & 0x0001)
				{
				KBD_D_PORT = KBD_D_PORT | (1 << KBD_D);
				}
			else
				{
				KBD_D_PORT = KBD_D_PORT & ~ (1 << KBD_D);
				}

			KBD_DATA = (KBD_DATA >> 1);
			}
		}
	}

int main(void)
	{
	UBRRL = MYUBR;
	UCSRB = (1 << TXEN);

	KBD_C_DDR = KBD_C_DDR & ~ (1 << KBD_C);					// CLOCK INPUT
	KBD_D_DDR = KBD_D_DDR & ~ (1 << KBD_D);					// DATA INPUT
	KBD_C_PORT = KBD_C_PORT | (1 << KBD_C);					// CLOCK PULLUP
	KBD_D_PORT = KBD_D_PORT | (1 << KBD_D);					// DATA PULLUP

	CRT_C_DDR = CRT_C_DDR & ~ (1 << CRT_C);					// CLOCK INPUT
	CRT_D_DDR = CRT_D_DDR & ~ (1 << CRT_D);					// DATA INPUT
	CRT_C_PORT = CRT_C_PORT | (1 << CRT_C);					// CLOCK PULLUP
	CRT_D_PORT = CRT_D_PORT | (1 << CRT_D);					// DATA PULLUP

	MCUCR = MCUCR | (1 << ISC11);						// The falling edge of INT1 generates an interrupt request.
	GICR = GICR | (1 << INT1);						// External Interrupt Request 1 Enable

	KBD_BIT_COUNT	= 11;
	KBD_DIR		= KBD_DIR_RECEIVE;

	sei();


	DDRD = DDRD | (1 << PD4);						// BUTTON INPUT
	PORTD = PORTD | (1 << PD4);						// BUTTON PULLUP

	while(1)
		{
		if(crt_get_state() == 0x02)					// CRT hat etwas zu melden
			{
			CRT_DATA = 0x0000;

			crt_read_frame();

			if(CRT_DATA != 0x0000)
				{
				usart_write_char('[');
				usart_write_hex_char((CRT_DATA >> 1) & 0xFF);
				usart_write_char(']');

				cli();

				KBD_DIR = KBD_DIR_SEND;

				KBD_C_PORT = KBD_C_PORT & ~ (1 << KBD_C);	// CLOCK OFF
				_delay_us(100);
				KBD_D_PORT = KBD_D_PORT & ~ (1 << KBD_D);	// DATA OFF
				KBD_C_PORT = KBD_C_PORT | (1 << KBD_C);		// CLOCK ON

				GIFR = (1 << INT0);

				sei();

				CRT_DATA = 0x0000;
				}
			}

		if((PIND & (1 << PD4)) != (1 << PD4))
			{
			crt_write_string("Hier die Daten der Magnetstreifenkarte senden ...\r\n");

			while((PIND & (1 << PD4)) != (1 << PD4));
			}

		}
	}

